/////////////////////////////////////////////////////////////////////////////////
// bh1792.c
// 
// Copyright (c) 2016-17 ROHM Co.,Ltd.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
/////////////////////////////////////////////////////////////////////////////////

#include "app.h"
#include "bh1792.h"


extern APP_DATA appData;

static uint32_t ReadCoreTimer(void);

static uint32_t ReadCoreTimer() {
    volatile uint32_t timer;

    // get the count reg
    asm volatile("mfc0   %0, $9" : "=r"(timer));

    return (timer);
}

#define GetSystemClock() (SYS_TIME_CPU_CLOCK_FREQUENCY)
#define us_SCALE   (GetSystemClock()/2000000)
#define ms_SCALE   (GetSystemClock()/2000)

/* Address of slave devices */
#define SLAVE_ADDRESS 0x5b

#define FORBIDDEN_PERIOD 195117//195215-195117
// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************

/* I2C Driver TX buffer  */
uint8_t TXbuffer[10];
/* I2C Driver RX buffer  */
uint8_t RXbuffer[10];

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Functions
// *****************************************************************************
// *****************************************************************************



// *****************************************************************************
// *****************************************************************************
// Section: Application Local Functions
// *****************************************************************************
// *****************************************************************************

void DelayMs(unsigned long int msDelay);
void DelayUs(unsigned long int usDelay);

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/* State Machine for Master Write */
bool APP_Read_Tasks(uint8_t);
bool APP_Write_Tasks(uint8_t);

// Global Variables

/* I2C Driver Handle  */
APP_I2C_DRIVER_HANDLE drvI2CHandle_Master;
APP_I2C_BUFFER_HANDLE drvI2CTxRxBufferHandle[MAX_NUMBER_OF_BUFFERS];

// Static Variables
static bh1792_t *s_pBH1792;

// Local Function Prototypes
static int32_t bh1792_getDataOut(bh1792_data_t *dat);
static int32_t bh1792_getFifoData(bh1792_data_t *dat);


/////////////////////////////////////////////////////////////////////////////////
//  Public Functions
/////////////////////////////////////////////////////////////////////////////////

//===============================================================================
// @brief  BH1792 Initialization
//
// @param[in]    : None
// @param[out]   : None
// @param[inout] : bh1792_t *pBH1792
// @retval       : BH1792_SUCCESS
//                 BH1792_I2C_ERR
//                 BH1792_NOT_EXIST
//===============================================================================

int32_t bh1792_Init(bh1792_t *pBH1792) {

    // Register the pointer of bh1792_t variable
    s_pBH1792 = pBH1792;

    // Set inital register parameters
    s_pBH1792->prm.sel_adc = BH1792_PRM_SEL_ADC_GREEN; // LED1, LED2 hCo 
    s_pBH1792->prm.msr = BH1792_PRM_MSR_32HZ; // 32Hz [h
    s_pBH1792->prm.led_en = BH1792_PRM_LED_EN1_0; //@pX@LED1ALED2ALED3 
    s_pBH1792->prm.led_cur1 = BH1792_PRM_LED_CUR1_MIN; // LED1ALED2@~[h
    s_pBH1792->prm.led_cur2 = BH1792_PRM_LED_CUR2_MIN; // LED3@~[h
    s_pBH1792->prm.ir_th = 0xFFFC; //@IR 荞臒l@default value 0xFFFC
    s_pBH1792->prm.int_sel = BH1792_PRM_INT_SEL_DISABLE; //@荞ݏo͂Ȃ

    // Clear flags
    s_pBH1792->i2c_err = 0;
    s_pBH1792->is_measuring = 0;
    s_pBH1792->sync_seq = 0;

    TXbuffer[0] = BH1792_ADDR_PARTID;
    APP_Read_Tasks(1);

    if (RXbuffer[0] == BH1792_PRM_PARTID) {
        
        TXbuffer[0] = BH1792_ADDR_RESET;
        TXbuffer[1] = BH1792_PRM_SWRESET << 7;
        APP_Write_Tasks(2U);
        
        return BH1792_SUCCESS;
        
    } else {
        return BH1792_NOT_EXIST;
    }
}

//===============================================================================
// @brief  BH1792 Parameter Setting
//
// @param[in]    : None
// @param[out]   : None
// @param[inout] : None
// @retval       : BH1792_SUCCESS
//                 BH1792_I2C_ERR
//===============================================================================

int32_t bh1792_SetParams(void) {

    static uint8_t TX_length = 7U;

    TXbuffer[0] = BH1792_ADDR_MEAS_CTRL1;

    // MEAS_CTRLx register
    TXbuffer[1] = (BH1792_PRM_RDY << 7) | (BH1792_PRM_SEL_LED << 6) | (s_pBH1792->prm.sel_adc << 4) | s_pBH1792->prm.msr;
    TXbuffer[2] = ((s_pBH1792->prm.led_en >> 1) << 6) | s_pBH1792->prm.led_cur1;
    TXbuffer[3] = ((s_pBH1792->prm.led_en & 0x01U) << 7) | s_pBH1792->prm.led_cur2;
    TXbuffer[4] = (uint8_t) s_pBH1792->prm.ir_th;
    TXbuffer[5] = (uint8_t) (s_pBH1792->prm.ir_th >> 8);
    TXbuffer[6] = s_pBH1792->prm.int_sel;
    // MEAS_START register
    if (s_pBH1792->is_measuring > 0) {
        TXbuffer[7] = BH1792_PRM_MEAS_ST;
        TX_length = 8U;
    }

    APP_Write_Tasks(TX_length);

    return BH1792_SUCCESS;
}

//===============================================================================
// @brief  BH1792 Measurement Start
//
// @param[in]    : None
// @param[out]   : None
// @param[inout] : None
// @retval       : BH1792_SUCCESS
//                 BH1792_I2C_ERR
//===============================================================================

int32_t bh1792_StartMeasure(void) {

    if (s_pBH1792->prm.msr < BH1792_PRM_MSR_SINGLE) {
        s_pBH1792->is_measuring = 1;

        // Initialize moving average parameters
        s_pBH1792->ma_prm.sum.off = 0U;
        s_pBH1792->ma_prm.sum.on = 0U;
        s_pBH1792->ma_prm.pos = 0;
        s_pBH1792->ma_prm.num = 0;

        switch (s_pBH1792->prm.msr) {
            case BH1792_PRM_MSR_32HZ:
            case BH1792_PRM_MSR_128HZ:
            case BH1792_PRM_MSR_64HZ:
                s_pBH1792->ma_prm.len = 1;
                break;
            case BH1792_PRM_MSR_256HZ:
                s_pBH1792->ma_prm.len = 2;
                break;
            case BH1792_PRM_MSR_1024HZ:
                s_pBH1792->ma_prm.len = 8;
                break;
            default:
                s_pBH1792->ma_prm.len = 0;
                break;
        }
    }
    
    TXbuffer[0] = BH1792_ADDR_MEAS_START;
    TXbuffer[1] = BH1792_PRM_MEAS_ST;
    APP_Write_Tasks(2);

    return BH1792_SUCCESS;

}

//===============================================================================
// @brief  BH1792 Measurement Stop
//
// @param[in]    : None
// @param[out]   : None
// @param[inout] : None
// @retval       : BH1792_SUCCESS
//                 BH1792_I2C_ERR
//===============================================================================

int32_t bh1792_StopMeasure(void) {

    s_pBH1792->is_measuring = 0;
    s_pBH1792->sync_seq = 0;

    TXbuffer[0] = BH1792_ADDR_RESET;
    TXbuffer[1] = BH1792_PRM_SWRESET << 7;
    APP_Write_Tasks(2);

    return BH1792_SUCCESS;
}

//===============================================================================
// @brief  BH1792 Measurement Data
//
// @param[in]    : None
// @param[out]   : bh1792_data_t *dat
// @param[inout] : None
// @retval       : BH1792_SUCCESS
//                 BH1792_I2C_ERR
//                 BH1792_FIFO_FULL
//===============================================================================

int32_t bh1792_GetMeasData(bh1792_data_t * dat) {
    static int32_t ret;
//    static uint8_t int_clear = 0U;
//    static uint8_t fifo_level = 0U;

    if (s_pBH1792->prm.msr <= BH1792_PRM_MSR_1024HZ) {
        TXbuffer[0] = BH1792_ADDR_FIFO_LEV;
        APP_Read_Tasks(1);

        dat->fifo_lev = RXbuffer[0];

        if (dat->fifo_lev == BH1792_PRM_FIFO_LEV_FULL) {
            ret = BH1792_FIFO_FULL;
        }

        if (s_pBH1792->prm.int_sel == BH1792_PRM_INT_SEL_WTM) {
            dat->fifo_lev = BH1792_PRM_FIFO_LEV_WTM;
        }

        bh1792_getFifoData(dat);

        TXbuffer[0] = BH1792_ADDR_FIFO_LEV;
        APP_Read_Tasks(1);
//        fifo_level = RXbuffer[0];

    } else {
        bh1792_getDataOut(dat);

        if (s_pBH1792->prm.int_sel >= BH1792_PRM_INT_SEL_IR) {
            TXbuffer[0] = BH1792_ADDR_INT_CLEAR;
            APP_Read_Tasks(1);
            ret = BH1792_SUCCESS;
        }
        return ret;
    }
          return ret;//????
}

//===============================================================================
// @brief  BH1792 Synchronization
//
// @param[in]    : None
// @param[out]   : None
// @param[inout] : None
// @retval       : BH1792_SUCCESS
//                 BH1792_I2C_ERR
//                 BH1792_NOT_EXIST
//                 BH1792_FIFO_FULL
//===============================================================================

int32_t bh1792_SetSync(void) {

    TXbuffer[0] = BH1792_ADDR_MEAS_SYNC;
    TXbuffer[1] = BH1792_PRM_MEAS_SYNC;
    APP_Write_Tasks(2);

    if (s_pBH1792->sync_seq < 3) {
        s_pBH1792->sync_seq++;
    }

    return BH1792_SUCCESS;

}

//===============================================================================
// @brief  BH1792 FIFO DATA Clear
//
// @param[in]    : None
// @param[out]   : None
// @param[inout] : None
// @retval       : BH1792_SUCCESS
//                 BH1792_I2C_ERR
//===============================================================================

int32_t bh1792_ClearFifoData(void) {

    static uint8_t i;
    static uint8_t r_cnt = 35U;
//    static uint8_t fifo_level = 0U;


    if (s_pBH1792->prm.msr == BH1792_PRM_MSR_32HZ) {
        r_cnt = 32U;
    }

    for (i = 0; i < r_cnt; i++) {
        TXbuffer[0] = BH1792_ADDR_FIFO_DATA0_LSBS;
        APP_Read_Tasks(4);
    }

    TXbuffer[0] = BH1792_ADDR_FIFO_LEV;
    APP_Read_Tasks(1);
//    fifo_level = RXbuffer[0];

    return BH1792_SUCCESS;
}

/////////////////////////////////////////////////////////////////////////////////
//  Local Functions
/////////////////////////////////////////////////////////////////////////////////

//===============================================================================
// @brief  BH1792 IR/GREEN DATA
//
// @param[in]    : None
// @param[out]   : bh1792_data_t *dat
// @param[inout] : None
// @retval       : I2C Error Code
//===============================================================================

int32_t bh1792_getDataOut(bh1792_data_t * dat) {

    uint8_t idx = 0U;

    if (s_pBH1792->prm.sel_adc == BH1792_PRM_SEL_ADC_GREEN) {
        idx = 4U;
    }

    TXbuffer[0] = BH1792_ADDR_IRDATA_LEDOFF_LSBS + idx;
    APP_Read_Tasks(4);

    dat->ir.off = ((uint16_t) RXbuffer[1] << 8) | (uint16_t) RXbuffer[0];
    dat->ir.on = ((uint16_t) RXbuffer[3] << 8) | (uint16_t) RXbuffer[2];
    dat->green.off = ((uint16_t) RXbuffer[5] << 8) | (uint16_t) RXbuffer[4];
    dat->green.on = ((uint16_t) RXbuffer[7] << 8) | (uint16_t) RXbuffer[6];

    return BH1792_SUCCESS;
}


//===============================================================================
// @brief  BH1792 FIFO DATA
//
// @param[in]    : None
// @param[out]   : bh1792_data_t *dat
// @param[inout] : None
// @retval       : I2C Error Code
//===============================================================================

int32_t bh1792_getFifoData(bh1792_data_t * dat) {
    static uint8_t i;
    static bh1792_maPrm_t *ma_prm;

    ma_prm = &s_pBH1792->ma_prm;

    TXbuffer[0] = BH1792_ADDR_FIFO_DATA0_LSBS;

    for (i = 0; i < dat->fifo_lev; i++) {

        APP_Read_Tasks(4);

        dat->fifo[i].off = ((uint16_t) RXbuffer[1] << 8) | (uint16_t) RXbuffer[0];
        dat->fifo[i].on = ((uint16_t) RXbuffer[3] << 8) | (uint16_t) RXbuffer[2];

        // Moving Average
        if (ma_prm->num == ma_prm->len) {
            ma_prm->sum.off -= ma_prm->buf[ma_prm->pos].off;
            ma_prm->sum.on -= ma_prm->buf[ma_prm->pos].on;
        }
        ma_prm->sum.off += dat->fifo[i].off;
        ma_prm->sum.on += dat->fifo[i].on;
        ma_prm->buf[ma_prm->pos].off = dat->fifo[i].off;
        ma_prm->buf[ma_prm->pos].on = dat->fifo[i].on;

        ma_prm->pos++;
        if (ma_prm->pos >= ma_prm->len) {
            ma_prm->pos = 0;
        }
        if (ma_prm->num < ma_prm->len) {
            ma_prm->num++;
        }

        if (ma_prm->num > 0) {
            dat->fifo_lpf[i].off = ma_prm->sum.off / ma_prm->num;
            dat->fifo_lpf[i].on = ma_prm->sum.on / ma_prm->num;
        }
    }
    return BH1792_SUCCESS;

}

//*********************************************
// I2C read write functions
//*********************************************

bool APP_Read_Tasks(uint8_t RX_length1) {

    if (s_pBH1792->prm.msr <= BH1792_PRM_MSR_1024HZ) {
        while (TMR2_CounterGet() >= FORBIDDEN_PERIOD); // I2CʐM֎~ԑт
    }

    I2C1_WriteRead(SLAVE_ADDRESS, &TXbuffer[0], 1, &RXbuffer[0], RX_length1);
     while(I2C1_IsBusy());
      
    return true;
}

bool APP_Write_Tasks(uint8_t TX_length) {

    if (s_pBH1792->prm.msr <= BH1792_PRM_MSR_1024HZ) {
        if ((TXbuffer[0] != BH1792_ADDR_MEAS_SYNC)) {
            while (TMR2_CounterGet() >= FORBIDDEN_PERIOD); // I2CʐM֎~ԑт
        }
    }

    I2C1_Write(SLAVE_ADDRESS, &TXbuffer[0], TX_length);
    while(I2C1_IsBusy());
    
    return true;
}

//****************************************************************************/
//  Function: APP_Check_Transfer_Status
//
//  Returns the status of Buffer operation from the driver; The application
//  can probe this function to see if the status of a particular I2C
//  transfer is in its execution
//****************************************************************************/



//****************************************************************************/
//  Function: Master Callback Function 
//
//  Callback from DRV_I2C_Tasks when I2C is configured in Master mode     
//****************************************************************************/


/***********************************************************
 *   Millisecond Delay function using the Count register
 *   in coprocessor 0 in the MIPS core.
 *   When running 200 MHz, CoreTimer frequency is 100 MHz
 *   CoreTimer increments every 2 SYS_CLK, CoreTimer period = 10ns
 *   1 ms = N x CoreTimer_period;
 *   To count 1ms, N = 100000 counts of CoreTimer
 *   1 ms = 10 ns * 100000 = 10e6 ns = 1 ms
 *   When running 80 MHz, CoreTimer frequency is 40 MHz 
 *   CoreTimer increments every 2 SYS_CLK, CoreTimer period = 25ns
 *   To count 1ms, N = 40000 counts of CoreTimer
 *   1ms = 25 ns * 40000 = 10e6 ns = 1 ms
 *   ms_SCALE = (GetSystemClock()/2000) @ 200 MHz = 200e6/2e3 = 100e3 = 100000
 *   ms_SCLAE = (GetSystemClock()/2000) @ = 80e6/2e3 = 40e3 = 40000 
 */

void DelayMs(unsigned long int msDelay) {
    register unsigned int startCntms = ReadCoreTimer();
    register unsigned int waitCntms = msDelay * ms_SCALE;

    while (ReadCoreTimer() - startCntms < waitCntms);
}

/***********************************************************
 *   Microsecond Delay function using the Count register
 *   in coprocessor 0 in the MIPS core.
 *   When running 200 MHz, CoreTimer frequency is 100 MHz
 *   CoreTimer increments every 2 SYS_CLK, CoreTimer period = 10ns
 *   1 us = N x CoreTimer_period;
 *   To count 1us, N = 100 counts of CoreTimer
 *   1 us = 10 ns * 100 = 1000 ns  = 1us
 *   When running 80 MHz, CoreTimer frequency is 40 MHz 
 *   CoreTimer increments every 2 SYS_CLK, CoreTimer period = 25ns
 *   To count 1us, N = 40 counts of CoreTimer
 *   1us = 25 ns * 40 = 1000 ns = 1 us
 *   us_SCALE = (GetSystemClock()/2000) @ 200 MHz = 200e6/2e6 = 100 
 *   us_SCLAE = (GetSystemClock()/2000) @ 80 MHz = 80e6/2e6 = 40 
 */

void DelayUs(unsigned long int usDelay) {
    register unsigned int startCnt = ReadCoreTimer();
    register unsigned int waitCnt = usDelay * us_SCALE;

    while (ReadCoreTimer() - startCnt < waitCnt);
}


